@@ -31,16 +31,16 @@ package org.jf.baksmali.Adaptors.Format;
 import org.jf.baksmali.Adaptors.MethodDefinition;
 import org.jf.baksmali.Adaptors.MethodDefinition.InvalidSwitchPayload;
 import org.jf.baksmali.Adaptors.MethodItem;
-import org.jf.baksmali.Renderers.LongRenderer;
+import org.jf.baksmali.Adaptors.ReferenceFormatter;
 import org.jf.baksmali.BaksmaliOptions;
+import org.jf.baksmali.Renderers.LongRenderer;
 import org.jf.dexlib2.Opcode;
-import org.jf.dexlib2.ReferenceType;
 import org.jf.dexlib2.VerificationError;
-import org.jf.dexlib2.dexbacked.DexBackedDexFile.InvalidItemIndex;
 import org.jf.dexlib2.iface.instruction.*;
 import org.jf.dexlib2.iface.instruction.formats.Instruction20bc;
 import org.jf.dexlib2.iface.instruction.formats.Instruction31t;
 import org.jf.dexlib2.iface.instruction.formats.UnknownInstruction;
+import org.jf.dexlib2.iface.reference.CallSiteReference;
 import org.jf.dexlib2.iface.reference.Reference;
 import org.jf.dexlib2.util.ReferenceUtil;
 import org.jf.util.ExceptionWithContext;
@@ -79,20 +79,16 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
         return opcode.isVolatileFieldAccessor() || opcode == Opcode.THROW_VERIFICATION_ERROR;
     }
 
-    private String writeInvalidItemIndex(InvalidItemIndex ex, int type, IndentingWriter writer)
-            throws IOException {
-        writer.write("#");
-        writer.write(ex.getMessage());
-        writer.write("\n");
-        return String.format("%s@%d", ReferenceType.toString(type), ex.getInvalidIndex());
+    private interface Writable {
+        void writeTo(IndentingWriter writer) throws IOException;
     }
 
     @Override
     public boolean writeTo(IndentingWriter writer) throws IOException {
         Opcode opcode = instruction.getOpcode();
         String verificationErrorName = null;
-        String referenceString = null;
-        String referenceString2 = null;
+        Writable referenceWritable = null;
+        Writable referenceWritable2 = null;
 
         boolean commentOutInstruction = false;
 
@@ -109,25 +105,41 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
 
         if (instruction instanceof ReferenceInstruction) {
             ReferenceInstruction referenceInstruction = (ReferenceInstruction)instruction;
-            String classContext = null;
+            final String classContext;
             if (methodDef.classDef.options.implicitReferences) {
                 classContext = methodDef.method.getDefiningClass();
+            } else {
+                classContext = null;
             }
 
-            try {
-                Reference reference = referenceInstruction.getReference();
-                referenceString = ReferenceUtil.getReferenceString(reference, classContext);
-                assert referenceString != null;
-            } catch (InvalidItemIndex ex) {
-                commentOutInstruction = true;
-                referenceString = writeInvalidItemIndex(ex, referenceInstruction.getReferenceType(),
-                        writer);
-            } catch (ReferenceType.InvalidReferenceTypeException ex) {
-                writer.write("#invalid reference type: ");
-                writer.printSignedIntAsDec(ex.getReferenceType());
-                commentOutInstruction = true;
+            Reference reference = referenceInstruction.getReference();
 
-                referenceString = "invalid_reference";
+            try {
+                reference.validateReference();
+
+                if (reference instanceof CallSiteReference) {
+                    referenceWritable = new Writable() {
+                        @Override
+                        public void writeTo(IndentingWriter indentingWriter) throws IOException {
+                            ReferenceFormatter.writeCallSiteReference(indentingWriter, (CallSiteReference)reference);
+                        }
+                    };
+                } else {
+                    referenceWritable = new Writable() {
+                        @Override
+                        public void writeTo(IndentingWriter indentingWriter) throws IOException {
+                            indentingWriter.write(ReferenceUtil.getReferenceString(reference, classContext));
+                        }
+                    };
+                }
+            } catch (Reference.InvalidReferenceException ex) {
+                commentOutInstruction = true;
+                writer.write("#");
+                writer.write(ex.getMessage());
+                writer.write("\n");
+                referenceWritable = indentingWriter -> {
+                    indentingWriter.write(ex.getInvalidReferenceRepresentation());
+                };
             }
 
             if (instruction instanceof DualReferenceInstruction) {
@@ -135,17 +147,19 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                         (DualReferenceInstruction) instruction;
                 try {
                     Reference reference2 = dualReferenceInstruction.getReference2();
-                    referenceString2 = ReferenceUtil.getReferenceString(reference2, classContext);
-                } catch (InvalidItemIndex ex) {
-                    commentOutInstruction = true;
-                    referenceString2 = writeInvalidItemIndex(ex,
-                            dualReferenceInstruction.getReferenceType2(), writer);
-                } catch (ReferenceType.InvalidReferenceTypeException ex) {
-                    writer.write("#invalid reference type: ");
-                    writer.printSignedIntAsDec(ex.getReferenceType());
-                    commentOutInstruction = true;
+                    reference2.validateReference();
 
-                    referenceString2 = "invalid_reference";
+                    referenceWritable2 = indentingWriter -> {
+                        indentingWriter.write(ReferenceUtil.getReferenceString(reference2, classContext));
+                    };
+                } catch (Reference.InvalidReferenceException ex) {
+                    commentOutInstruction = true;
+                    writer.write("#");
+                    writer.write(ex.getMessage());
+                    writer.write("\n");
+                    referenceWritable = indentingWriter -> {
+                        indentingWriter.write(ex.getInvalidReferenceRepresentation());
+                    };
                 }
             }
         }
@@ -235,7 +249,7 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(' ');
                 writer.write(verificationErrorName);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 break;
             case Format20t:
             case Format30t:
@@ -249,7 +263,7 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(' ');
                 writeFirstRegister(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 break;
             case Format21ih:
             case Format21lh:
@@ -293,7 +307,7 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(", ");
                 writeSecondRegister(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 break;
             case Format22cs:
                 writeOpcode(writer);
@@ -335,7 +349,7 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(' ');
                 writeInvokeRegisters(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 break;
             case Format35mi:
                 writeOpcode(writer);
@@ -356,7 +370,7 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(' ');
                 writeInvokeRangeRegisters(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 break;
             case Format3rmi:
                 writeOpcode(writer);
@@ -377,18 +391,18 @@ public class InstructionMethodItem<T extends Instruction> extends MethodItem {
                 writer.write(' ');
                 writeInvokeRegisters(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 writer.write(", ");
-                writer.write(referenceString2);
+                referenceWritable2.writeTo(writer);
                 break;
             case Format4rcc:
                 writeOpcode(writer);
                 writer.write(' ');
                 writeInvokeRangeRegisters(writer);
                 writer.write(", ");
-                writer.write(referenceString);
+                referenceWritable.writeTo(writer);
                 writer.write(", ");
-                writer.write(referenceString2);
+                referenceWritable2.writeTo(writer);
                 break;
             default:
                 assert false;
