@@ -473,6 +473,9 @@ public class MutableMethodImplementation implements MethodImplementation {
     private int mapCodeAddressToIndex(@Nonnull int[] codeAddressToIndex, int codeAddress) {
         int index;
         do {
+            if (codeAddress >= codeAddressToIndex.length) {
+                codeAddress = codeAddressToIndex.length - 1;
+            }
             index = codeAddressToIndex[codeAddress];
             if (index < 0) {
                 codeAddress--;
@@ -601,6 +604,9 @@ public class MutableMethodImplementation implements MethodImplementation {
             case Format22c:
                 setInstruction(location, newBuilderInstruction22c((Instruction22c) instruction));
                 return;
+            case Format22cs:
+                setInstruction(location, newBuilderInstruction22cs((Instruction22cs) instruction));
+                return;
             case Format22s:
                 setInstruction(location, newBuilderInstruction22s((Instruction22s) instruction));
                 return;
@@ -615,9 +621,6 @@ public class MutableMethodImplementation implements MethodImplementation {
             case Format23x:
                 setInstruction(location, newBuilderInstruction23x((Instruction23x) instruction));
                 return;
-            case Format25x:
-                setInstruction(location, newBuilderInstruction25x((Instruction25x) instruction));
-                return;
             case Format30t:
                 setInstruction(location, newBuilderInstruction30t(location.codeAddress,
                         codeAddressToIndex,
@@ -639,9 +642,21 @@ public class MutableMethodImplementation implements MethodImplementation {
             case Format35c:
                 setInstruction(location, newBuilderInstruction35c((Instruction35c) instruction));
                 return;
+            case Format35mi:
+                setInstruction(location, newBuilderInstruction35mi((Instruction35mi) instruction));
+                return;
+            case Format35ms:
+                setInstruction(location, newBuilderInstruction35ms((Instruction35ms) instruction));
+                return;
             case Format3rc:
                 setInstruction(location, newBuilderInstruction3rc((Instruction3rc)instruction));
                 return;
+            case Format3rmi:
+                setInstruction(location, newBuilderInstruction3rmi((Instruction3rmi)instruction));
+                return;
+            case Format3rms:
+                setInstruction(location, newBuilderInstruction3rms((Instruction3rms)instruction));
+                return;
             case Format51l:
                 setInstruction(location, newBuilderInstruction51l((Instruction51l)instruction));
                 return;
@@ -773,6 +788,15 @@ public class MutableMethodImplementation implements MethodImplementation {
                 instruction.getReference());
     }
 
+    @Nonnull
+    private BuilderInstruction22cs newBuilderInstruction22cs(@Nonnull Instruction22cs instruction) {
+        return new BuilderInstruction22cs(
+                instruction.getOpcode(),
+                instruction.getRegisterA(),
+                instruction.getRegisterB(),
+                instruction.getFieldOffset());
+    }
+
     @Nonnull
     private BuilderInstruction22s newBuilderInstruction22s(@Nonnull Instruction22s instruction) {
         return new BuilderInstruction22s(
@@ -872,15 +896,29 @@ public class MutableMethodImplementation implements MethodImplementation {
     }
 
     @Nonnull
-    private BuilderInstruction25x newBuilderInstruction25x(@Nonnull Instruction25x instruction) {
-        return new BuilderInstruction25x(
+    private BuilderInstruction35mi newBuilderInstruction35mi(@Nonnull Instruction35mi instruction) {
+        return new BuilderInstruction35mi(
                 instruction.getOpcode(),
-                instruction.getParameterRegisterCount(),
-                instruction.getRegisterFixedC(),
-                instruction.getRegisterParameterD(),
-                instruction.getRegisterParameterE(),
-                instruction.getRegisterParameterF(),
-                instruction.getRegisterParameterG());
+                instruction.getRegisterCount(),
+                instruction.getRegisterC(),
+                instruction.getRegisterD(),
+                instruction.getRegisterE(),
+                instruction.getRegisterF(),
+                instruction.getRegisterG(),
+                instruction.getInlineIndex());
+    }
+
+    @Nonnull
+    private BuilderInstruction35ms newBuilderInstruction35ms(@Nonnull Instruction35ms instruction) {
+        return new BuilderInstruction35ms(
+                instruction.getOpcode(),
+                instruction.getRegisterCount(),
+                instruction.getRegisterC(),
+                instruction.getRegisterD(),
+                instruction.getRegisterE(),
+                instruction.getRegisterF(),
+                instruction.getRegisterG(),
+                instruction.getVtableIndex());
     }
 
     @Nonnull
@@ -892,6 +930,24 @@ public class MutableMethodImplementation implements MethodImplementation {
                 instruction.getReference());
     }
 
+    @Nonnull
+    private BuilderInstruction3rmi newBuilderInstruction3rmi(@Nonnull Instruction3rmi instruction) {
+        return new BuilderInstruction3rmi(
+                instruction.getOpcode(),
+                instruction.getStartRegister(),
+                instruction.getRegisterCount(),
+                instruction.getInlineIndex());
+    }
+
+    @Nonnull
+    private BuilderInstruction3rms newBuilderInstruction3rms(@Nonnull Instruction3rms instruction) {
+        return new BuilderInstruction3rms(
+                instruction.getOpcode(),
+                instruction.getStartRegister(),
+                instruction.getRegisterCount(),
+                instruction.getVtableIndex());
+    }
+
     @Nonnull
     private BuilderInstruction51l newBuilderInstruction51l(@Nonnull Instruction51l instruction) {
         return new BuilderInstruction51l(
