@@ -40,12 +40,7 @@ import org.jf.dexlib2.iface.instruction.DualReferenceInstruction;
 import org.jf.dexlib2.iface.instruction.ReferenceInstruction;
 import org.jf.dexlib2.iface.instruction.SwitchElement;
 import org.jf.dexlib2.iface.instruction.formats.*;
-import org.jf.dexlib2.iface.reference.FieldReference;
-import org.jf.dexlib2.iface.reference.MethodProtoReference;
-import org.jf.dexlib2.iface.reference.MethodReference;
-import org.jf.dexlib2.iface.reference.Reference;
-import org.jf.dexlib2.iface.reference.StringReference;
-import org.jf.dexlib2.iface.reference.TypeReference;
+import org.jf.dexlib2.iface.reference.*;
 import org.jf.util.ExceptionWithContext;
 
 import javax.annotation.Nonnull;
@@ -55,7 +50,8 @@ import java.util.List;
 
 public class InstructionWriter<StringRef extends StringReference, TypeRef extends TypeReference,
         FieldRefKey extends FieldReference, MethodRefKey extends MethodReference,
-        ProtoRefKey extends MethodProtoReference> {
+        ProtoRefKey extends MethodProtoReference, MethodHandleKey extends MethodHandleReference,
+        CallSiteKey extends CallSiteReference> {
     @Nonnull private final Opcodes opcodes;
     @Nonnull private final DexDataWriter writer;
     @Nonnull private final StringSection<?, StringRef> stringSection;
@@ -63,10 +59,14 @@ public class InstructionWriter<StringRef extends StringReference, TypeRef extend
     @Nonnull private final FieldSection<?, ?, FieldRefKey, ?> fieldSection;
     @Nonnull private final MethodSection<?, ?, ?, MethodRefKey, ?> methodSection;
     @Nonnull private final ProtoSection<?, ?, ProtoRefKey, ?> protoSection;
+    @Nonnull private final MethodHandleSection<MethodHandleKey, ?, ?> methodHandleSection;
+    @Nonnull private final CallSiteSection<CallSiteKey, ?> callSiteSection;
 
-    @Nonnull static <StringRef extends StringReference, TypeRef extends TypeReference, FieldRefKey extends FieldReference,
-            MethodRefKey extends MethodReference, ProtoRefKey extends MethodProtoReference>
-            InstructionWriter<StringRef, TypeRef, FieldRefKey, MethodRefKey, ProtoRefKey>
+    @Nonnull static <StringRef extends StringReference, TypeRef extends TypeReference,
+            FieldRefKey extends FieldReference, MethodRefKey extends MethodReference,
+            ProtoRefKey extends MethodProtoReference, MethodHandleKey extends MethodHandleReference,
+            CallSiteKey extends CallSiteReference>
+            InstructionWriter<StringRef, TypeRef, FieldRefKey, MethodRefKey, ProtoRefKey, MethodHandleKey, CallSiteKey>
             makeInstructionWriter(
                 @Nonnull Opcodes opcodes,
                 @Nonnull DexDataWriter writer,
@@ -74,9 +74,13 @@ public class InstructionWriter<StringRef extends StringReference, TypeRef extend
                 @Nonnull TypeSection<?, ?, TypeRef> typeSection,
                 @Nonnull FieldSection<?, ?, FieldRefKey, ?> fieldSection,
                 @Nonnull MethodSection<?, ?, ?, MethodRefKey, ?> methodSection,
-                @Nonnull ProtoSection<?, ?, ProtoRefKey, ?> protoSection) {
-        return new InstructionWriter<StringRef, TypeRef, FieldRefKey, MethodRefKey, ProtoRefKey>(
-                opcodes, writer, stringSection, typeSection, fieldSection, methodSection, protoSection);
+                @Nonnull ProtoSection<?, ?, ProtoRefKey, ?> protoSection,
+                @Nonnull MethodHandleSection<MethodHandleKey, ?, ?> methodHandleSection,
+                @Nonnull CallSiteSection<CallSiteKey, ?> callSiteSection) {
+        return new InstructionWriter<
+                StringRef, TypeRef, FieldRefKey, MethodRefKey, ProtoRefKey, MethodHandleKey,CallSiteKey>(
+                        opcodes, writer, stringSection, typeSection, fieldSection, methodSection, protoSection,
+                        methodHandleSection, callSiteSection);
     }
 
     InstructionWriter(@Nonnull Opcodes opcodes,
@@ -85,7 +89,9 @@ public class InstructionWriter<StringRef extends StringReference, TypeRef extend
                       @Nonnull TypeSection<?, ?, TypeRef> typeSection,
                       @Nonnull FieldSection<?, ?, FieldRefKey, ?> fieldSection,
                       @Nonnull MethodSection<?, ?, ?, MethodRefKey, ?> methodSection,
-                      @Nonnull ProtoSection<?, ?, ProtoRefKey, ?> protoSection) {
+                      @Nonnull ProtoSection<?, ?, ProtoRefKey, ?> protoSection,
+                      @Nonnull MethodHandleSection<MethodHandleKey, ?, ?> methodHandleSection,
+                      @Nonnull CallSiteSection<CallSiteKey, ?> callSiteSection) {
         this.opcodes = opcodes;
         this.writer = writer;
         this.stringSection = stringSection;
@@ -93,6 +99,8 @@ public class InstructionWriter<StringRef extends StringReference, TypeRef extend
         this.fieldSection = fieldSection;
         this.methodSection = methodSection;
         this.protoSection = protoSection;
+        this.methodHandleSection = methodHandleSection;
+        this.callSiteSection = callSiteSection;
     }
 
     private short getOpcodeValue(Opcode opcode) {
@@ -551,6 +559,10 @@ public class InstructionWriter<StringRef extends StringReference, TypeRef extend
                 return typeSection.getItemIndex((TypeRef) reference);
             case ReferenceType.METHOD_PROTO:
                 return protoSection.getItemIndex((ProtoRefKey) reference);
+            case ReferenceType.METHOD_HANDLE:
+                return methodHandleSection.getItemIndex((MethodHandleKey) reference);
+            case ReferenceType.CALL_SITE:
+                return callSiteSection.getItemIndex((CallSiteKey) reference);
             default:
                 throw new ExceptionWithContext("Unknown reference type: %d",  referenceType);
         }
