@@ -35,6 +35,7 @@ import com.google.common.base.Function;
 import com.google.common.base.Predicate;
 import com.google.common.collect.*;
 import org.jf.dexlib2.DebugItemType;
+import org.jf.dexlib2.HiddenApiRestriction;
 import org.jf.dexlib2.builder.MutableMethodImplementation;
 import org.jf.dexlib2.iface.ExceptionHandler;
 import org.jf.dexlib2.iface.Field;
@@ -48,9 +49,9 @@ import org.jf.dexlib2.iface.value.EncodedValue;
 import org.jf.dexlib2.util.EncodedValueUtils;
 import org.jf.dexlib2.writer.ClassSection;
 import org.jf.dexlib2.writer.DebugWriter;
+import org.jf.dexlib2.writer.builder.BuilderEncodedValues.BuilderArrayEncodedValue;
 import org.jf.dexlib2.writer.builder.BuilderEncodedValues.BuilderEncodedValue;
 import org.jf.util.AbstractForwardSequentialList;
-import org.jf.util.CollectionUtils;
 import org.jf.util.ExceptionWithContext;
 
 import javax.annotation.Nonnull;
@@ -62,7 +63,7 @@ import java.util.concurrent.ConcurrentMap;
 
 public class BuilderClassPool extends BaseBuilderPool implements ClassSection<BuilderStringReference,
         BuilderTypeReference, BuilderTypeList, BuilderClassDef, BuilderField, BuilderMethod, BuilderAnnotationSet,
-        BuilderEncodedValue> {
+        BuilderArrayEncodedValue> {
     @Nonnull private final ConcurrentMap<String, BuilderClassDef> internedItems =
             Maps.newConcurrentMap();
 
@@ -153,24 +154,8 @@ public class BuilderClassPool extends BaseBuilderPool implements ClassSection<Bu
             };
 
     @Nullable @Override
-    public Collection<? extends BuilderEncodedValue> getStaticInitializers(@Nonnull BuilderClassDef classDef) {
-        final SortedSet<BuilderField> sortedStaticFields = classDef.getStaticFields();
-
-        final int lastIndex = CollectionUtils.lastIndexOf(sortedStaticFields, HAS_INITIALIZER);
-        if (lastIndex > -1) {
-            return new AbstractCollection<BuilderEncodedValue>() {
-                @Nonnull @Override public Iterator<BuilderEncodedValue> iterator() {
-                    return FluentIterable.from(sortedStaticFields)
-                            .limit(lastIndex+1)
-                            .transform(GET_INITIAL_VALUE).iterator();
-                }
-
-                @Override public int size() {
-                    return lastIndex+1;
-                }
-            };
-        }
-        return null;
+    public BuilderArrayEncodedValue getStaticInitializers(@Nonnull BuilderClassDef classDef) {
+        return classDef.staticInitializers;
     }
 
     @Nonnull @Override
@@ -211,6 +196,16 @@ public class BuilderClassPool extends BaseBuilderPool implements ClassSection<Bu
         return builderMethod.accessFlags;
     }
 
+    @Nonnull @Override
+    public Set<HiddenApiRestriction> getFieldHiddenApiRestrictions(@Nonnull BuilderField builderField) {
+        return builderField.getHiddenApiRestrictions();
+    }
+
+    @Nonnull @Override
+    public Set<HiddenApiRestriction> getMethodHiddenApiRestrictions(@Nonnull BuilderMethod builderMethod) {
+        return builderMethod.getHiddenApiRestrictions();
+    }
+
     @Nullable @Override public BuilderAnnotationSet getClassAnnotations(@Nonnull BuilderClassDef builderClassDef) {
         if (builderClassDef.annotations.isEmpty()) {
             return null;
@@ -325,14 +320,6 @@ public class BuilderClassPool extends BaseBuilderPool implements ClassSection<Bu
         return new MutableMethodImplementation(impl);
     }
 
-    @Override public void setEncodedArrayOffset(@Nonnull BuilderClassDef builderClassDef, int offset) {
-        builderClassDef.encodedArrayOffset = offset;
-    }
-
-    @Override public int getEncodedArrayOffset(@Nonnull BuilderClassDef builderClassDef) {
-        return builderClassDef.encodedArrayOffset;
-    }
-
     @Override public void setAnnotationDirectoryOffset(@Nonnull BuilderClassDef builderClassDef, int offset) {
         builderClassDef.annotationDirectoryOffset = offset;
     }
