@@ -33,8 +33,12 @@ package org.jf.dexlib2.writer.pool;
 
 import com.google.common.base.Function;
 import com.google.common.base.Predicate;
-import com.google.common.collect.*;
+import com.google.common.collect.FluentIterable;
+import com.google.common.collect.ImmutableList;
+import com.google.common.collect.Iterables;
+import com.google.common.collect.Ordering;
 import org.jf.dexlib2.DebugItemType;
+import org.jf.dexlib2.HiddenApiRestriction;
 import org.jf.dexlib2.ReferenceType;
 import org.jf.dexlib2.builder.MutableMethodImplementation;
 import org.jf.dexlib2.iface.*;
@@ -42,14 +46,13 @@ import org.jf.dexlib2.iface.debug.*;
 import org.jf.dexlib2.iface.instruction.Instruction;
 import org.jf.dexlib2.iface.instruction.ReferenceInstruction;
 import org.jf.dexlib2.iface.reference.*;
+import org.jf.dexlib2.iface.value.ArrayEncodedValue;
 import org.jf.dexlib2.iface.value.EncodedValue;
-import org.jf.dexlib2.immutable.value.ImmutableEncodedValueFactory;
-import org.jf.dexlib2.util.EncodedValueUtils;
 import org.jf.dexlib2.util.ReferenceUtil;
 import org.jf.dexlib2.writer.ClassSection;
 import org.jf.dexlib2.writer.DebugWriter;
+import org.jf.dexlib2.writer.util.StaticInitializerUtil;
 import org.jf.util.AbstractForwardSequentialList;
-import org.jf.util.CollectionUtils;
 import org.jf.util.ExceptionWithContext;
 
 import javax.annotation.Nonnull;
@@ -60,7 +63,7 @@ import java.util.Map.Entry;
 
 public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSection<CharSequence, CharSequence,
         TypeListPool.Key<? extends Collection<? extends CharSequence>>, PoolClassDef, Field, PoolMethod,
-        Set<? extends Annotation>, EncodedValue> {
+        Set<? extends Annotation>, ArrayEncodedValue> {
 
     public ClassPool(@Nonnull DexPool dexPool) {
         super(dexPool);
@@ -94,6 +97,11 @@ public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSe
             }
 
             dexPool.annotationSetSection.intern(field.getAnnotations());
+
+            ArrayEncodedValue staticInitializers = getStaticInitializers(poolClassDef);
+            if (staticInitializers != null) {
+                dexPool.encodedArraySection.intern(staticInitializers);
+            }
         }
 
         HashSet<String> methods = new HashSet<String>();
@@ -139,6 +147,9 @@ public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSe
                         case ReferenceType.METHOD:
                             dexPool.methodSection.intern((MethodReference)reference);
                             break;
+                        case ReferenceType.CALL_SITE:
+                            dexPool.callSiteSection.intern((CallSiteReference) reference);
+                            break;
                         default:
                             throw new ExceptionWithContext("Unrecognized reference type: %d",
                                     instruction.getOpcode().referenceType);
@@ -240,44 +251,9 @@ public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSe
         return classDef.getSourceFile();
     }
 
-    private static final Predicate<Field> HAS_INITIALIZER = new Predicate<Field>() {
-        @Override
-        public boolean apply(Field input) {
-            EncodedValue encodedValue = input.getInitialValue();
-            return encodedValue != null && !EncodedValueUtils.isDefaultValue(encodedValue);
-        }
-    };
-
-    private static final Function<Field, EncodedValue> GET_INITIAL_VALUE = new Function<Field, EncodedValue>() {
-        @Override
-        public EncodedValue apply(Field input) {
-            EncodedValue initialValue = input.getInitialValue();
-            if (initialValue == null) {
-                return ImmutableEncodedValueFactory.defaultValueForType(input.getType());
-            }
-            return initialValue;
-        }
-    };
-
-    @Nullable @Override public Collection<? extends EncodedValue> getStaticInitializers(
+    @Nullable @Override public ArrayEncodedValue getStaticInitializers(
             @Nonnull PoolClassDef classDef) {
-        final SortedSet<Field> sortedStaticFields = classDef.getStaticFields();
-
-        final int lastIndex = CollectionUtils.lastIndexOf(sortedStaticFields, HAS_INITIALIZER);
-        if (lastIndex > -1) {
-            return new AbstractCollection<EncodedValue>() {
-                @Nonnull @Override public Iterator<EncodedValue> iterator() {
-                    return FluentIterable.from(sortedStaticFields)
-                            .limit(lastIndex+1)
-                            .transform(GET_INITIAL_VALUE).iterator();
-                }
-
-                @Override public int size() {
-                    return lastIndex+1;
-                }
-            };
-        }
-        return null;
+        return StaticInitializerUtil.getStaticInitializers(classDef.getStaticFields());
     }
 
     @Nonnull @Override public Collection<? extends Field> getSortedStaticFields(@Nonnull PoolClassDef classDef) {
@@ -312,6 +288,14 @@ public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSe
         return method.getAccessFlags();
     }
 
+    @Nonnull @Override public Set<HiddenApiRestriction> getFieldHiddenApiRestrictions(@Nonnull Field field) {
+        return field.getHiddenApiRestrictions();
+    }
+
+    @Nonnull @Override public Set<HiddenApiRestriction> getMethodHiddenApiRestrictions(@Nonnull PoolMethod poolMethod) {
+        return poolMethod.getHiddenApiRestrictions();
+    }
+
     @Nullable @Override public Set<? extends Annotation> getClassAnnotations(@Nonnull PoolClassDef classDef) {
         Set<? extends Annotation> annotations = classDef.getAnnotations();
         if (annotations.size() == 0) {
@@ -421,14 +405,6 @@ public class ClassPool extends BasePool<String, PoolClassDef> implements ClassSe
         return new MutableMethodImplementation(poolMethod.getImplementation());
     }
 
-    @Override public void setEncodedArrayOffset(@Nonnull PoolClassDef classDef, int offset) {
-        classDef.encodedArrayOffset = offset;
-    }
-
-    @Override public int getEncodedArrayOffset(@Nonnull PoolClassDef classDef) {
-        return classDef.encodedArrayOffset;
-    }
-
     @Override public void setAnnotationDirectoryOffset(@Nonnull PoolClassDef classDef, int offset) {
         classDef.annotationDirectoryOffset = offset;
     }
